from dataclasses import dataclass
from .model import Model


@dataclass
class ControlsController:
    model: Model

    def up_action(self) -> None:
        self.model.bpm += 1

    def down_action(self) -> None:
        self.model.bpm -= 1


@dataclass
class PlayPauseController:
    model: Model

    def button_action(self) -> None:
        if self.model.is_playing:
            self.model.stop()
        else:
            self.model.start()



from typing import Generic, Protocol, TypeVar

T = TypeVar('T')
S = TypeVar('S', contravariant=True)


class Observer(Protocol[S]):
    def update_(self, value: S) -> None:
        ...


class Observable(Generic[T]):
    def __init__(self, observers: str) -> None:
        self.observers = observers

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    def __set__(self, instance: object, value: T) -> None:
        instance.__dict__[self.name] = value
        observers = instance.__dict__[self.observers]
        for observer in observers:
            observer.update_(value)

    def __get__(self, instance: object, owner: type = None) -> T:
        return instance.__dict__[self.name]



import threading
from time import sleep
from types import TracebackType
from typing import Callable, Optional, Type

from .observable import Observable, Observer


class Model:
    bpm = Observable[int]('_bpm_observers')
    is_playing = Observable[bool]('_is_playing_observers')

    def __init__(self, bpm: int, beat: Callable[[], None]) -> None:
        self._bpm_observers: list[Observer[int]] = []
        self.bpm = bpm
        self._beat = beat
        self._initialize_thread()
        self._is_playing_observers: list[Observer[bool]] = []
        self.is_playing = False

    def __enter__(self) -> None:
        self.start()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> None:
        self.quit()

    def register_for_bpm(self, observer: Observer[int]) -> None:
        self._bpm_observers.append(observer)

    def register_for_is_playing(self, observer: Observer[bool]) -> None:
        self._is_playing_observers.append(observer)

    @property
    def _seconds_per_beat(self) -> float:
        return 60 / self.bpm

    def _initialize_thread(self) -> None:
        self._can_play = threading.Event()
        self._quit_event = threading.Event()
        beat_thread = threading.Thread(target=self._beat_loop)
        beat_thread.start()

    def start(self) -> None:
        self._can_play.set()
        self.is_playing = True

    def stop(self) -> None:
        self._can_play.clear()
        self.is_playing = False

    def quit(self) -> None:
        self._quit_event.set()
        self._can_play.set()  # cannot quit when paused

    def _beat_loop(self) -> None:
        while not self._quit_event.is_set():
            self._beat()
            sleep(self._seconds_per_beat)
            self._can_play.wait()



import tkinter as tk

from .controller import ControlsController, PlayPauseController
from .model import Model


class Display(tk.Frame):
    def __init__(self, master: tk.Frame, model: Model) -> None:
        super().__init__(master=master, height=2)

        self.bpm = tk.Label(master=self, text=str(model.bpm))
        self.bpm.pack(side=tk.LEFT)

        unit = tk.Label(master=self, text='bpm')
        unit.pack(side=tk.RIGHT)

        model.register_for_bpm(self)

    def update_(self, value: int) -> None:
        self.bpm['text'] = str(value)


class Controls(tk.Frame):
    def __init__(
        self,
        master: tk.Frame,
        controller: ControlsController,
    ) -> None:
        super().__init__(master=master)

        up_button = tk.Button(master=self, text='+', width=2)
        up_button.bind('<Button 1>', lambda event: controller.up_action())
        up_button.pack()

        down_button = tk.Button(master=self, text='-', width=2)
        down_button.bind('<Button 1>', lambda event: controller.down_action())
        down_button.pack()


class PlayPauseView(tk.Frame):
    def __init__(
        self,
        master: tk.Frame,
        model: Model,
        controller: PlayPauseController,
    ) -> None:
        super().__init__(master=master)

        self.button = tk.Button(master=self, text='play')
        self.button.bind(
            '<Button 1>',
            lambda event: controller.button_action()
        )
        self.button.pack()

        model.register_for_is_playing(self)
        self.update_(model.is_playing)

    def update_(self, value: bool) -> None:
        self.button['text'] = 'pause' if value else 'play'


class View(tk.Frame):
    def __init__(self, master: tk.Tk, model: Model) -> None:
        super().__init__(master=master)

        left_frame = tk.Frame(master=self)
        display = Display(master=left_frame, model=model)
        display.pack(side=tk.TOP)

        play_pause_controller = PlayPauseController(model)
        play_pause = PlayPauseView(
            master=left_frame,
            model=model,
            controller=play_pause_controller
        )
        play_pause.pack(side=tk.BOTTOM)

        left_frame.pack(side=tk.LEFT)

        controls_controller = ControlsController(model=model)
        controls = Controls(
            master=self,
            controller=controls_controller,
        )
        controls.pack(side=tk.RIGHT)

        self.pack()



from argparse import ArgumentParser
from importlib import resources
import tkinter as tk

from . import assets
from .model import Model
from .sound import play_sound
from .view import View


def get_bpm() -> int:
    parser = ArgumentParser()
    parser.add_argument('bpm', type=int, default=120)
    return parser.parse_args().bpm


def run_metronome(bpm: int) -> None:
    with resources.path(assets, 'click.wav') as path:
        model = Model(bpm=bpm, beat=lambda: play_sound(path))
        with model:
            window = tk.Tk()
            view = View(master=window, model=model)
            window.mainloop()


def main() -> None:
    bpm = get_bpm()
    run_metronome(bpm)


if __name__ == '__main__':
    main()



import pathlib

from setuptools import setup


HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(
    name='etronome',
    version='0.0.1',
    description='Keep track of a rythm',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/wpbindt/metronome',
    author='Wessel Bindt',
    author_email='wesselbindt@gmail.com',
    license='GNU',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.8'
    ],
    packages=['etronome'],
    include_package_data=True,
)

